# Copyright 2023-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
`dataskech` (https://github.com/ekzhu/datasketch) doesn't offer a way to store the hash tables in disk. This
is a custom implementation that uses `shelve` to store the hash tables in disk.
Note: This implementation is not optimized for performance, but could be worth
creating a PR to `datasketch`.
"""

import shutil
import struct
from pathlib import Path
from typing import Callable, Dict, Final, Optional, Tuple

from datasketch import MinHashLSH as _MinHashLSH
from datasketch.lsh import _optimal_param
from datasketch.storage import OrderedStorage, UnorderedStorage, _random_name
from datasketch.storage import ordered_storage as _ordered_storage
from datasketch.storage import unordered_storage as _unordered_storage

KEY_VALUE_DISK_DIR: Path = Path.home() / ".cache" / "distilabel" / "key_value_store"
KV_DISK_LIST_NAME: Final[str] = "disckache_list_storage"
KV_DISK_SET_NAME: Final[str] = "diskcache_set_storage"


class DiskCacheListStorage(OrderedStorage):
    def __init__(self, config, name) -> None:
        path = config.get("path", self._get_db_name(name))
        try:
            from diskcache import Index
        except ImportError as e:
            raise ImportError(
                "`diskcache` is required for disk storage using `MinHashDedup`. "
                "Please install it using `pip install diskcache`."
            ) from e

        # Start with a clean file on each pipeline
        if Path(path).exists():
            shutil.rmtree(path)
        self._db = Index(path)

    def _get_db_name(self, name):
        return str(KEY_VALUE_DISK_DIR / f"{name}_{KV_DISK_LIST_NAME}")

    def keys(self):
        return self._db.keys()

    def get(self, key):
        return self._db.get(key, [])

    def remove(self, *keys):
        self._db.clear()

    def remove_val(self, key, val):
        self.get(key).remove(val)

    def insert(self, key, *vals, **kwargs):
        res = self.get(key)
        res.extend(vals)
        self._db[key] = res

    def size(self):
        return len(self._db)

    def itemcounts(self, **kwargs):
        return {k: len(v) for k, v in self._db.items()}

    def has_key(self, key):
        return key in self._db

    def close(self):
        self._db._cache.close()


class DiskCacheSetStorage(UnorderedStorage, DiskCacheListStorage):
    def _get_db_name(self, name):
        return str(KEY_VALUE_DISK_DIR / f"{name}_{KV_DISK_SET_NAME}")

    def get(self, key):
        return self._db.get(key, set())

    def insert(self, key, *vals, **kwargs):
        res = self.get(key)
        res.update(vals)
        self._db[key] = res


def ordered_storage(config, name=None):
    """Copy of `datasketch.storage.ordered_storage` with the addition of `ShelveListStorage`."""
    tp = config["type"]
    if tp == "disk":
        return DiskCacheListStorage(config, name=name)
    return _ordered_storage(config, name=name)


def unordered_storage(config, name=None):
    """Copy of `datasketch.storage.ordered_storage` with the addition of `ShelveSetStorage`."""
    tp = config["type"]
    if tp == "disk":
        return DiskCacheSetStorage(config, name=name)
    return _unordered_storage(config, name=name)


class MinHashLSH(_MinHashLSH):
    """Custom implementation of `datasketch.MinHashLSH` to allow passing a custom
    storage configuration to store the hash tables in disk.

    This could be merged in the original repository, the only changes
    to the __init__ are the additional `close` method, and the use
    of our custom `ordered_storage` and `unordered_storage` functions.
    """

    def __init__(
        self,
        threshold: float = 0.9,
        num_perm: int = 128,
        weights: Tuple[float, float] = (0.5, 0.5),
        params: Optional[Tuple[int, int]] = None,
        storage_config: Optional[Dict] = None,
        prepickle: Optional[bool] = None,
        hashfunc: Optional[Callable[[bytes], bytes]] = None,
    ) -> None:
        storage_config = {"type": "dict"} if not storage_config else storage_config
        self._buffer_size = 50000
        if threshold > 1.0 or threshold < 0.0:
            raise ValueError("threshold must be in [0.0, 1.0]")
        if num_perm < 2:
            raise ValueError("Too few permutation functions")
        if any(w < 0.0 or w > 1.0 for w in weights):
            raise ValueError("Weight must be in [0.0, 1.0]")
        if sum(weights) != 1.0:
            raise ValueError("Weights must sum to 1.0")
        self.h = num_perm
        if params is not None:
            self.b, self.r = params
            if self.b * self.r > num_perm:
                raise ValueError(
                    "The product of b and r in params is "
                    "{} * {} = {} -- it must be less than num_perm {}. "
                    "Did you forget to specify num_perm?".format(
                        self.b, self.r, self.b * self.r, num_perm
                    )
                )
        else:
            false_positive_weight, false_negative_weight = weights
            self.b, self.r = _optimal_param(
                threshold, num_perm, false_positive_weight, false_negative_weight
            )
        if self.b < 2:
            raise ValueError("The number of bands are too small (b < 2)")

        self.prepickle = (
            storage_config["type"] == "redis" if not prepickle else prepickle
        )

        self.hashfunc = hashfunc
        if hashfunc:
            self._H = self._hashed_byteswap
        else:
            self._H = self._byteswap

        basename = storage_config.get("basename", _random_name(11))
        self.hashtables = [
            unordered_storage(
                storage_config,
                name=b"".join([basename, b"_bucket_", struct.pack(">H", i)]),
            )
            for i in range(self.b)
        ]
        self.hashranges = [(i * self.r, (i + 1) * self.r) for i in range(self.b)]
        self.keys = ordered_storage(storage_config, name=b"".join([basename, b"_keys"]))

    def close(self):
        """Closes the internal connections."""
        if isinstance(self.hashtables[0], DiskCacheListStorage):
            for ht in self.hashtables:
                ht.close()
            self.keys.close()



# Copyright 2023-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
from typing import TYPE_CHECKING, Any, List, Set

from pydantic import BaseModel, Field

from distilabel.utils.serialization import TYPE_INFO_KEY

if TYPE_CHECKING:
    pass

# Add here the name of the attributes that shouldn't be used to generate the signature.
# Attributes from a `BaseModel` that is an attribute from the root class must be prefixed
# with the name of the attribute followed by an underscore. For example, if the attribute
# `jobs_ids` is an attribute from the `llm` attribute of the root class it should be added
# as `llm_jobs_ids`.
_EXCLUDE_FROM_SIGNATURE_DEFAULTS = {
    TYPE_INFO_KEY,
    "disable_cuda_device_placement",
    "input_batch_size",
    "gpu_memory_utilization",
    "resources",
    "exclude_from_signature",
    "llm_jobs_ids",
    "llm_offline_batch_generation_block_until_done",
}


class SignatureMixin(BaseModel):
    """Mixin for creating a signature (for cache) of the class.

    Attributes:
        exclude_from_signature: list of attributes to exclude from the signature.
    """

    exclude_from_signature: Set[str] = Field(
        default=_EXCLUDE_FROM_SIGNATURE_DEFAULTS, exclude=True
    )

    @property
    def signature(self) -> str:
        """Makes a signature (hash) of the class, using its attributes.

        Returns:
            signature of the class.
        """

        def flatten_dump(d: Any, parent_key: str = "", sep: str = "_") -> List:
            items = []
            for k, v in d.items():
                new_key = parent_key + sep + k if parent_key else k
                if isinstance(v, dict):
                    items.extend(flatten_dump(v, new_key, sep=sep))
                elif isinstance(v, list):
                    if len(v) == 0:
                        items.append((new_key, ""))
                    elif isinstance(v[0], str):
                        items.append((new_key, "-".join(v)))
                    else:
                        for i, x in enumerate(v):
                            items.extend(flatten_dump(x, f"{new_key}-{i}", sep=sep))
                elif new_key not in self.exclude_from_signature:
                    items.append((new_key, v))
            return items

        info = []
        for name, value in flatten_dump(self.dump()):
            info.append(f"{name}-{str(value)}")

        return hashlib.sha1("-".join(info).encode()).hexdigest()



# Copyright 2023-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import TYPE_CHECKING

import pytest

from distilabel.steps.clustering.text_clustering import TextClustering
from tests.unit.conftest import DummyAsyncLLM

if TYPE_CHECKING:
    from distilabel.llms.typing import GenerateOutput
    from distilabel.steps.tasks.typing import FormattedInput


class ClusteringLLM(DummyAsyncLLM):
    n: int = 1

    async def agenerate(  # type: ignore
        self, input: "FormattedInput", num_generations: int = 1
    ) -> "GenerateOutput":
        if self.n == 1:
            return [json.dumps({"labels": "label"}) for _ in range(num_generations)]
        return [
            json.dumps({"labels": ["label" for _ in range(self.n)]})
            for _ in range(self.n)
        ]


class TestTextClustering:
    @pytest.mark.parametrize("n", [1, 3])
    def test_process(self, n: int) -> None:
        step = TextClustering(
            llm=ClusteringLLM(n=n),
            n=n,
            samples_per_cluster=2,
            savefig=False,
        )
        step.load()

        results = next(
            step.process(
                inputs=[
                    {"projection": [0.1, -0.4], "cluster_label": -1, "text": "hello"},
                    {"projection": [-0.3, 0.9], "cluster_label": -1, "text": "hello"},
                    {"projection": [0.6, 0.2], "cluster_label": 0, "text": "hello"},
                    {"projection": [-0.2, -0.6], "cluster_label": 0, "text": "hello"},
                    {"projection": [0.9, 0.1], "cluster_label": 0, "text": "hello"},
                    {"projection": [0.4, -0.7], "cluster_label": 1, "text": "hello"},
                    {"projection": [-0.5, 0.3], "cluster_label": 1, "text": "hello"},
                    {"projection": [0.7, 0.5], "cluster_label": 2, "text": "hello"},
                    {"projection": [-0.1, -0.9], "cluster_label": 2, "text": "hello"},
                ]
            )
        )
        for r in results:
            if r["cluster_label"] == -1:
                assert r["summary_label"] == json.dumps("Unclassified")
            else:
                if n == 1:
                    assert r["summary_label"] == json.dumps("label")
                else:
                    assert r["summary_label"] == json.dumps(["label"] * n)
